#include <sys/time.h>
#include <signal.h>
#include <unistd.h>
#include <stdio.h>

static long p_realt_secs = 0, c1_realt_secs = 0, c2_realt_secs = 0;
static long p_virtt_secs = 0, c1_virtt_secs = 0, c2_virtt_secs = 0;
static long p_proft_secs = 0, c1_proft_secs = 0, c2_proft_secs = 0;
static struct itimerval p_realt, c1_realt, c2_realt;
static struct itimerval p_virtt, c1_virtt, c2_virtt;
static struct itimerval p_proft, c1_proft, c2_proft;

long unsigned int fibonacci(unsigned int n)
{
	if(n == 0)
		return 0;
	else if((n == 1) || (n == 2))
		return 1;
	else
		return(fibonacci(n-1) + fibonacci(n-2));
}

void p_signal(int sig)
{
	switch(sig)
	{
	case SIGALRM:
		p_realt_secs++;
		break;
	case SIGVTALRM:
		p_virtt_secs++;
		break;
	case SIGPROF:
		p_proft_secs++;
		break;
	default:
		break;
	}
	
	return;
}

void c1_signal(int sig)
{
	switch(sig)
	{
	case SIGALRM:
		c1_realt_secs++;
		break;
	case SIGVTALRM:
		c1_virtt_secs++;
		break;
	case SIGPROF:
		c1_proft_secs++;
		break;
	default:
		break;
	}
	
	return;	
}

void c2_signal(int sig)
{
	switch(sig)
	{
	case SIGALRM:
		c2_realt_secs++;
		break;
	case SIGVTALRM:
		c2_virtt_secs++;
		break;
	case SIGPROF:
		c2_proft_secs++;
		break;
	default:
		break;
	}
	
	return;
}

long unsigned int elapsed_usecs(long sec, long usec)
{
	if( sec == 1)
		return 0;
	else
		return (1000000 - usec);
}

long unsigned delta_time(long p_sec, long v_sec)
{
	return (p_sec - v_sec);
}

main(int argc, char **argv)
{
	long unsigned fib = 0;
	int pid1, pid2;
	unsigned int fibarg;
	int status;
	
	//Get command line argument, fibarg
	fibarg = atoi( argv[1] );
	
	//Initialize parent, child1, child2 timer values
	p_realt.it_interval.tv_sec = 1;
	p_realt.it_interval.tv_usec = 0;
	p_realt.it_value.tv_sec = 1;
	p_realt.it_value.tv_usec = 0;

	p_virtt.it_interval.tv_sec = 1;
	p_virtt.it_interval.tv_usec = 0;
	p_virtt.it_value.tv_sec = 1;
	p_virtt.it_value.tv_usec = 0;

	p_proft.it_interval.tv_sec = 1;
	p_proft.it_interval.tv_usec = 0;
	p_proft.it_value.tv_sec = 1;
	p_proft.it_value.tv_usec = 0;

	c1_realt.it_interval.tv_sec = 1;
	c1_realt.it_interval.tv_usec = 0;
	c1_realt.it_value.tv_sec = 1;
	c1_realt.it_value.tv_usec = 0;

	c1_virtt.it_interval.tv_sec = 1;
	c1_virtt.it_interval.tv_usec = 0;
	c1_virtt.it_value.tv_sec = 1;
	c1_virtt.it_value.tv_usec = 0;

	c1_proft.it_interval.tv_sec = 1;
	c1_proft.it_interval.tv_usec = 0;
	c1_proft.it_value.tv_sec = 1;
	c1_proft.it_value.tv_usec = 0;

	c2_realt.it_interval.tv_sec = 1;
	c2_realt.it_interval.tv_usec = 0;
	c2_realt.it_value.tv_sec = 1;
	c2_realt.it_value.tv_usec = 0;

	c2_virtt.it_interval.tv_sec = 1;
	c2_virtt.it_interval.tv_usec = 0;
	c2_virtt.it_value.tv_sec = 1;
	c2_virtt.it_value.tv_usec = 0;

	c2_proft.it_interval.tv_sec = 1;
	c2_proft.it_interval.tv_usec = 0;
	c2_proft.it_value.tv_sec = 1;
	c2_proft.it_value.tv_usec = 0;

	//Enable parent's signal handlers
	signal(SIGALRM, p_signal);
	signal(SIGVTALRM, p_signal);
	signal(SIGPROF, p_signal);
	
	//Set parent's itimers
	if( setitimer(ITIMER_REAL, &p_realt, NULL) == -1)
		perror("parent real timer set error");
	if( setitimer(ITIMER_VIRTUAL, &p_virtt, NULL) == -1)
		perror("parent virtual timer set error");
	if( setitimer(ITIMER_PROF, &p_proft, NULL) == -1)
		perror("parent profile timer set error");
	
	pid1 = fork();
	
	if(pid1 == 0)
	{
		//Enable child 1 signal handlers
		signal(SIGALRM, c1_signal);
		signal(SIGVTALRM, c1_signal);
		signal(SIGPROF, c1_signal);
		
		//Set child 1 itimers
		if( setitimer(ITIMER_REAL, &c1_realt, NULL) == -1)
			perror("child 1 real timer set error");
		if( setitimer(ITIMER_VIRTUAL, &c1_virtt, NULL) == -1)
			perror("child 1 virtual timer set error");
		if( setitimer(ITIMER_PROF, &c1_proft, NULL) == -1)
			perror("child 1 profile timer set error");
		
		//Start child 1 on the fibonacci program
		fib = fibonacci(fibarg);
		
		//Read child 1 itimer values and report them
		getitimer(ITIMER_PROF, &c1_proft);
		getitimer(ITIMER_REAL, &c1_realt);
		getitimer(ITIMER_VIRTUAL, &c1_virtt);
		
		printf("\n");
		
		printf("Child 1 fib = %ld, real time = %ld sec, %ld msec\n",
		fib, c1_realt_secs,
		elapsed_usecs(c1_realt.it_value.tv_sec, c1_realt.it_value.tv_usec)/1000 );

		printf("Child 1 fib = %ld, cpu time = %ld sec, %ld msec\n",
		fib, c1_proft_secs,
		elapsed_usecs(c1_proft.it_value.tv_sec, c1_proft.it_value.tv_usec)/1000 );
		
		printf("Child 1 fib = %ld, user time = %ld sec, %ld msec\n",
		fib, c1_virtt_secs,
		elapsed_usecs(c1_virtt.it_value.tv_sec, c1_virtt.it_value.tv_usec)/1000 );
		
		printf("Child 1 fib = %ld, kernel time = %ld sec, %ld msec\n",
		fib, delta_time(c1_proft_secs, c1_virtt_secs),
		(elapsed_usecs(c1_proft.it_value.tv_sec, c1_proft.it_value.tv_usec)/1000 ) -
		(elapsed_usecs(c1_virtt.it_value.tv_sec, c1_virtt.it_value.tv_usec)/1000 ));
		
		fflush(stdout);
		
		exit(0);
	}
	else
	{
		pid2 = fork();
		
		if(pid2 == 0)
		{
			//Enable child 2 signal handlers
			signal(SIGALRM, c2_signal);
			signal(SIGVTALRM, c2_signal);
			signal(SIGPROF, c2_signal);
			
			//Set child 2 itimers
			if( setitimer(ITIMER_REAL, &c2_realt, NULL) == -1)
				perror("child 2 real timer set error");
			if( setitimer(ITIMER_VIRTUAL, &c2_virtt, NULL) == -1)
				perror("child 2 virtual timer set error");
			if( setitimer(ITIMER_PROF, &c2_proft, NULL) == -1)
				perror("child 2 profile timer set error");
				
			//Start child 2 on the fibonacci program
			fib = fibonacci(fibarg);
			
			//Read child 2 itimer values and report them
			getitimer(ITIMER_PROF, &c2_proft);
			getitimer(ITIMER_REAL, &c2_realt);
			getitimer(ITIMER_VIRTUAL, &c2_virtt);
			
			printf("\n");
		
			printf("Child 2 fib = %ld, real time = %ld sec, %ld msec\n",
			fib, c2_realt_secs,
			elapsed_usecs(c2_realt.it_value.tv_sec, c2_realt.it_value.tv_usec)/1000 );

			printf("Child 2 fib = %ld, cpu time = %ld sec, %ld msec\n",
			fib, c2_proft_secs,
			elapsed_usecs(c2_proft.it_value.tv_sec, c2_proft.it_value.tv_usec)/1000 );
		
			printf("Child 2 fib = %ld, user time = %ld sec, %ld msec\n",
			fib, c2_virtt_secs,
			elapsed_usecs(c2_virtt.it_value.tv_sec, c2_virtt.it_value.tv_usec)/1000 );
		
			printf("Child 2 fib = %ld, kernel time = %ld sec, %ld msec\n",
			fib, delta_time(c2_proft_secs, c2_virtt_secs),
			(elapsed_usecs(c2_proft.it_value.tv_sec, c2_proft.it_value.tv_usec)/1000 ) -
			(elapsed_usecs(c2_virtt.it_value.tv_sec, c2_virtt.it_value.tv_usec)/1000 ));
		
			fflush(stdout);
		
			exit(0);	
		}
		else
		{
			/* this is the parent */
			
			//Start parent on the fibonacci program
			fib = fibonacci(fibarg);
			
			//Wait for children to terminate
			waitpid(0, &status, 0);
			waitpid(0, &status, 0);
			
			//Read parent itimer values and report them
			getitimer(ITIMER_PROF, &p_proft);
			getitimer(ITIMER_REAL, &p_realt);
			getitimer(ITIMER_VIRTUAL, &p_virtt);
			
			printf("\n");
		
			printf("Parent fib = %ld, real time = %ld sec, %ld msec\n",
			fib, p_realt_secs,
			elapsed_usecs(p_realt.it_value.tv_sec, p_realt.it_value.tv_usec)/1000 );

			printf("Parent fib = %ld, cpu time = %ld sec, %ld msec\n",
			fib, p_proft_secs,
			elapsed_usecs(p_proft.it_value.tv_sec, p_proft.it_value.tv_usec)/1000 );
		
			printf("Parent fib = %ld, user time = %ld sec, %ld msec\n",
			fib, p_virtt_secs,
			elapsed_usecs(p_virtt.it_value.tv_sec, p_virtt.it_value.tv_usec)/1000 );
		
			printf("Parent fib = %ld, kernel time = %ld sec, %ld msec\n",
			fib, delta_time(p_proft_secs, p_virtt_secs),
			(elapsed_usecs(p_proft.it_value.tv_sec, p_proft.it_value.tv_usec)/1000 ) -
			(elapsed_usecs(p_virtt.it_value.tv_sec, p_virtt.it_value.tv_usec)/1000 ));
		
			fflush(stdout);
		
			exit(0);
		}
	}
	printf("this line should never be printed.\n");
}
